/*
** PHREL
** $Id: prefs.c,v 1.43 2013/06/01 04:55:31 sella Exp $
** Copyright (c) 2004-2013 James M. Sella. All Rights Reserved.
** Released under the GPL Version 2 License.
** http://www.digitalgenesis.com
*/

static const char rcsid[] __attribute__((used)) = "$Id: prefs.c,v 1.43 2013/06/01 04:55:31 sella Exp $";

#include "prefs.h"
#include "db.h"

#include "phrel.h"
#include "util.h"

#include <snmptrap.h>
#include <configuration.h>

#include <pcap.h>
#include <ctype.h>
#include <string.h>
#include <syslog.h>
#include <pthread.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <errno.h>
#include <assert.h>

/* Global variables */
extern struct args_t args;
extern int errno;

const char algo_names[MAX_ALGO_NAMES][4] = { "avg", "sum", "max" };
char const short_options[] = "I:46P:p:T:R:B:D:X:Y:A:i:c:s:tw:mj:x:y:S:H:K:N:U:W:MG:L:C:u:g:r:o::z::n:d::vh";
struct option const long_options[] = {
	{"interface", required_argument, NULL, 'I'},
	{"ipv4", no_argument, NULL, '4'},
	{"ipv6", no_argument, NULL, '6'},
	{"protocol", required_argument, NULL, 'P'},
	{"port", required_argument, NULL, 'p'},
	{"threshold", required_argument, NULL, 'T'},
	{"rate", required_argument, NULL, 'R'},
	{"burst", required_argument, NULL, 'B'},
	{"decay", required_argument, NULL, 'D'},
	{"exclude", required_argument, NULL, 'X'},
	{"include", required_argument, NULL, 'Y'},
	{"algo", required_argument, NULL, 'A'},
	{"interval", required_argument, NULL, 'i'},
	{"check-interval", required_argument, NULL, 'c'},
	{"stats-interval", required_argument, NULL, 's'},
	{"snmptrap", required_argument, NULL, 'S'},
	{"test", no_argument, NULL, 't'},
	{"direction", required_argument, NULL, 'w'},
	{"promiscuous", no_argument, NULL, 'm'},
	{"jump", required_argument, NULL, 'j'},
	{"iptables", required_argument, NULL, 'x'},
	{"ip6tables", required_argument, NULL, 'y'},
	{"dbhost", required_argument, NULL, 'H'},
	{"dbsock", required_argument, NULL, 'K'},
	{"dbname", required_argument, NULL, 'N'},
	{"dbuser", required_argument, NULL, 'U'},
	{"dbpass", required_argument, NULL, 'W'},
	{"dbcomp", no_argument, NULL, 'M'},
	{"dbalgo", required_argument, NULL, 'G'},
	{"dbthreshold", required_argument, NULL, 'L'},
	{"config", required_argument, NULL, 'C'},
	{"user", required_argument, NULL, 'u'},
	{"group", required_argument, NULL, 'g'},
	{"chroot", required_argument, NULL, 'r'},
	{"stats", optional_argument, NULL, 'o'},
	{"dump", optional_argument, NULL, 'z'},
	{"nice", required_argument, NULL, 'n'},
	{"debug", optional_argument, NULL, 'd'},
	{"version", no_argument, NULL, 'v'},
	{"help", no_argument, NULL, 'h'},
	{NULL, 0, NULL, 0}
};

int get_args(int argc, char** argv) {
	int c, i, len, warn = 0, opt_index = 0;
	char *ptr, buf[512];
	struct threshold_t **t;
	struct prefix_t **p;
#ifdef HAVE_NET_SNMP
	snmptrap_peer_t **s;
#endif
	struct in_addr ip;
	struct in6_addr ip6;

	/* Default our settings. */
	memset(&args, 0, sizeof(struct args_t));
	args.ipv4 = USE_DEFAULT;
	args.ipv6 = USE_DEFAULT;
	args.nice = DEF_NICENESS;
	args.rate = DEF_RATELIMIT;
	args.burst = DEF_BURST;
	args.decay = DEF_DECAY;
	args.algo = DEF_PPS_ALGO;
	args.interval = DEF_INTERVAL;
	args.check_interval = DEF_CHECK_INTERVAL;
	args.stats_interval = DEF_STATS_INTERVAL;
	args.stale_interval = (args.interval * STALE_INTERVAL_MLT);
	args.max_samples = (args.interval + 1);
	args.db_algo = DEF_DBALGO;
	args.direction = DEF_DIRECTION;

	/* Set passed arguments. */
	while ((c = getopt_long(argc, argv, short_options, long_options, &opt_index)) != -1) {
		switch (c) {
			case 'I':
				if (args.interface == NULL) {
					args.interface = (char*) strdup(optarg);
				}

				break;

			case '4':
				args.ipv4 = 1;

				break;

			case '6':
				args.ipv6 = 1;

				break;

			case 'P':
				if (args.expression == NULL) {
					snprintf(buf, sizeof(buf), "%s", optarg);
				} else {
					snprintf(buf, sizeof(buf), "(%s) or %s", args.expression, optarg);
					free(args.expression);
				}
				args.expression = strdup(buf);

				break;

			case 'p':
				if (args.expression == NULL) {
					snprintf(buf, sizeof(buf), "dst port %s", optarg);
				} else {
					snprintf(buf, sizeof(buf), "%s and dst port %s", args.expression, optarg);
					free(args.expression);
				}
				args.expression = strdup(buf);

				break;
				
			case 'T': /* --threshold=PPS[:RATE[:BURST[:DECAY]]] */
				/* FIXME: If thresholds are specified by the user non-ascending, do they get used?? */
				t = &(args.threshold);
				while ((*t) != NULL) t = &((*t)->next); /* Move to the end of threshold list */
				*t = malloc(sizeof(struct threshold_t));
				memset(*t, 0, sizeof(struct threshold_t));

				(*t)->threshold = strtoul(optarg, NULL, 10);

				if ((*t)->threshold > MAX_THRESHOLD) {
					(*t)->threshold = MAX_THRESHOLD;
				}

				if ((ptr = strchr(optarg, ':')) != NULL && strlen(ptr) > 1) { /* RATE */
					ptr++;
					(*t)->rate = strtoul(ptr, NULL, 10);

					if ((*t)->rate > MAX_RATELIMIT) {
						(*t)->rate = MAX_RATELIMIT;
					}

					if ((ptr = strchr(ptr, ':')) != NULL && strlen(ptr) > 1) { /* BURST */
						ptr++;
						(*t)->burst = strtoul(ptr, NULL, 10);

						if ((*t)->burst > MAX_BURST) {
							(*t)->burst = MAX_BURST;
						}

						if ((ptr = strchr(ptr, ':')) != NULL && strlen(ptr) > 1) { /* DECAY */
							ptr++;
							(*t)->decay = strtoul(ptr, NULL, 10);

							if ((*t)->decay > MAX_DECAY) {
								(*t)->decay = MAX_DECAY;
							}
						} else {
							(*t)->decay = USE_DEFAULT;
						}
					} else {
						(*t)->burst = USE_DEFAULT;
						(*t)->decay = USE_DEFAULT;
					}
				} else {
					(*t)->rate = USE_DEFAULT;
					(*t)->burst = USE_DEFAULT;
					(*t)->decay = USE_DEFAULT;
				}

				break;
				
			case 'R':
				args.rate = strtoul(optarg, NULL, 10);

				if (args.rate > MAX_RATELIMIT) {
					args.rate = MAX_RATELIMIT;
				}

				break;
				
			case 'B':
				args.burst = strtoul(optarg, NULL, 10);

				if (args.burst > MAX_BURST) {
					args.burst = MAX_BURST;
				}

				break;
				
			case 'D':
				args.decay = strtoul(optarg, NULL, 10);

				if (args.decay > MAX_DECAY) {
					args.decay = MAX_DECAY;
				}

				break;
				
			case 'X':
				p = &(args.exclude);
				while ((*p) != NULL) p = &((*p)->next); /* Move to the end of exclude list */
				*p = malloc(sizeof(struct prefix_t));
				memset(*p, 0, sizeof(struct prefix_t));

				if (strchr(optarg, ':') == NULL) {
					(*p)->ip_ss.ss_family = AF_INET;
					(*p)->mask_ss.ss_family = AF_INET;

					if ((ptr = strchr(optarg, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
						*ptr = '\0'; /* Separate mask from ip */
						ptr++; /* Move past seperator */

						(*p)->cidr = (strtoul(ptr, NULL, 10) < 32) ? strtoul(ptr, NULL, 10) : 32;
					} else {
						(*p)->cidr = 32;
					}

					cidr_to_mask_in_addr(&ip, (*p)->cidr);
					set_in_addr(&((*p)->mask_ss), &ip);

					if (inet_pton(AF_INET, optarg, &ip) == 0) {
						fprintf(stderr, "excluded prefix '%s' has an invalid format\n", optarg);

						return EXIT_FAILURE;
					}

					set_in_addr(&((*p)->ip_ss), &ip);
					ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);
				} else {
					/* IPv6 part */
					(*p)->ip_ss.ss_family = AF_INET6;
					(*p)->mask_ss.ss_family = AF_INET6;

					if ((ptr = strchr(optarg, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
						*ptr = '\0'; /* Separate mask from ip */
						ptr++; /* Move past seperator */

						(*p)->cidr = (strtoul(ptr, NULL, 10) < 128) ? strtoul(ptr, NULL, 10) : 128;
					} else {
						(*p)->cidr = 128;
					}

					cidr_to_mask_in6_addr(&ip6, (*p)->cidr);
					set_in6_addr(&((*p)->mask_ss), &ip6);

					if (inet_pton(AF_INET6, optarg, &ip6) == 0) {
						fprintf(stderr, "excluded prefix '%s' has an invalid format\n", optarg);

						return EXIT_FAILURE;
					}

					set_in6_addr(&((*p)->ip_ss), &ip6);
					ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);
				}

				break;
				
			case 'Y':
				p = &(args.include);
				while ((*p) != NULL) p = &((*p)->next); /* Move to the end of include list */
				*p = malloc(sizeof(struct prefix_t));
				memset(*p, 0, sizeof(struct prefix_t));

				if (strchr(optarg, ':') == NULL) {
					(*p)->ip_ss.ss_family = AF_INET;
					(*p)->mask_ss.ss_family = AF_INET;

					if ((ptr = strchr(optarg, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
						*ptr = '\0'; /* Separate mask from ip */
						ptr++; /* Move past seperator */

						(*p)->cidr = (strtoul(ptr, NULL, 10) < 32) ? strtoul(ptr, NULL, 10) : 32;
					} else {
						(*p)->cidr = 32;
					}

					cidr_to_mask_in_addr(&ip, (*p)->cidr);
					set_in_addr(&((*p)->mask_ss), &ip);

					if (inet_pton(AF_INET, optarg, &ip) == 0) {
						fprintf(stderr, "included prefix '%s' has an invalid format\n", optarg);

						return EXIT_FAILURE;
					}

					set_in_addr(&((*p)->ip_ss), &ip);
					ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);
				} else {
					/* IPv6 part */
					(*p)->ip_ss.ss_family = AF_INET6;
					(*p)->mask_ss.ss_family = AF_INET6;

					if ((ptr = strchr(optarg, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
						*ptr = '\0'; /* Separate mask from ip */
						ptr++; /* Move past seperator */

						(*p)->cidr = (strtoul(ptr, NULL, 10) < 128) ? strtoul(ptr, NULL, 10) : 128;
					} else {
						(*p)->cidr = 128;
					}

					cidr_to_mask_in6_addr(&ip6, (*p)->cidr);
					set_in6_addr(&((*p)->mask_ss), &ip6);

					if (inet_pton(AF_INET6, optarg, &ip6) == 0) {
						fprintf(stderr, "included prefix '%s' has an invalid format\n", optarg);

						return EXIT_FAILURE;
					}

					set_in6_addr(&((*p)->ip_ss), &ip6);
					ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);
				}

				break;
				
			case 'A':
				for (i = 0; i < MAX_ALGO_NAMES; i++) {
					if (strncasecmp(optarg, algo_names[i], strlen(algo_names[i])) == 0) {
						args.algo = i;

						break;
					}
				}

				break;
				
			case 'i':
				args.interval = strtoul(optarg, NULL, 10);

				if (args.interval > MAX_INTERVAL) {
					args.interval = MAX_INTERVAL;
				} else if (args.interval < MIN_INTERVAL) {
					args.interval = MIN_INTERVAL;
				}

				args.stale_interval = (args.interval * STALE_INTERVAL_MLT);
				args.max_samples = (args.interval + 1);

				break;
				
			case 'c':
				args.check_interval = strtoul(optarg, NULL, 10);

				if (args.check_interval > MAX_CHECK_INTERVAL) {
					args.check_interval = MAX_CHECK_INTERVAL;
				} else if (args.check_interval < MIN_CHECK_INTERVAL) {
					args.check_interval = MIN_CHECK_INTERVAL;
				}

				break;
				
			case 's':
				args.stats_interval = strtoul(optarg, NULL, 10);

				if (args.stats_interval > MAX_STATS_INTERVAL) {
					args.stats_interval = MAX_STATS_INTERVAL;
				} else if (args.stats_interval < MIN_STATS_INTERVAL) {
					args.stats_interval = MIN_STATS_INTERVAL;
				}

				break;
				
			case 'S': /* --snmptrap=HOST[:PORT][:COMMUNITY[:VERSION[:INFORM]]] */
#ifdef HAVE_NET_SNMP
				s = &(args.peer);
				while ((*s) != NULL) s = &((*s)->next); /* Move to the end of peer list */
				*s = malloc(sizeof(snmptrap_peer_t));
				memset(*s, 0, sizeof(snmptrap_peer_t));
				(*s)->version = (*s)->inform = -1;

				ptr = strtok(optarg, ":");
				while (ptr != NULL) {
					if ((*s)->peername[0] == '\0') { /* HOST */
						strncpy((*s)->peername, ptr, sizeof((*s)->peername));

						/* Check if PORT is following hostname */
						if ((ptr = strtok(NULL, ":")) != NULL) { 
							/* Assume it's a PORT if its only numbers */
							snprintf(buf, sizeof(buf), "%u", atoi(ptr));
							if (strcmp(ptr, buf) == 0) { /* PORT */
								if (atoi(ptr) >= 1 && atoi(ptr) <= 65535) { /* Ignore invalid range */
									snprintf(buf, sizeof(buf), "%s:%u", (*s)->peername, atoi(ptr));
									strncpy((*s)->peername, buf, sizeof((*s)->peername));
								} else if (warn == 0) {
									fprintf(stderr, "Specified port '%s' for '--snmptrap' is invalid and will be ignored.\n", ptr);
									warn = 1;
								}
							} else {
								continue;
							}
						}
					} else if ((*s)->community[0] == '\0') { /* COMMUNITY */
						strncpy((*s)->community, ptr, sizeof((*s)->community));
					} else if ((*s)->version == -1) { /* VERSION */
						if (strcasecmp(ptr, "1") == 0 || strcasecmp(ptr, "v1") == 0) {
							(*s)->version = SNMP_VERSION_1;
						} else if (strcasecmp(ptr, "3") == 0 || strcasecmp(ptr, "v3") == 0) {
							(*s)->version = SNMP_VERSION_3;
						} else {
							(*s)->version = SNMP_VERSION_2c;
						}
					} else if ((*s)->inform == -1) { /* INFORM */
						if (strncasecmp(ptr, "y", 1) == 0 || strcasecmp(ptr, "on") == 0 || strcasecmp(ptr, "1") == 0) { /* Yes, On or 1 */
							(*s)->inform = 1;
						} else {
							(*s)->inform = 0;
						}
					} else {
						break;
					}

					/* Next token */
					ptr = strtok(NULL, ":");
				}

				/* Set COMMUNITY default, if needed */
				if ((*s)->community[0] == '\0') {
					strncpy((*s)->community, "public", sizeof((*s)->community));
				}

				/* Set VERSION default, if needed */
				if ((*s)->version == -1) {
					(*s)->version = SNMP_VERSION_2c;
				}

				/* Set INFORM default, if needed */
				if ((*s)->inform == -1) {
					(*s)->inform = 0;
				}
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--snmptrap' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_NET_SNMP */

				break;
				
			case 't':
				args.test = 1;

				break;
				
			case 'w':
				if (strncasecmp(optarg, "inout", 5) == 0) {
					args.direction = PCAP_D_INOUT;
				} else if (strncasecmp(optarg, "out", 3) == 0) {
					args.direction = PCAP_D_OUT;
				} else if (strncasecmp(optarg, "in", 2) == 0) {
					args.direction = PCAP_D_IN;
				}

				break;
				
			case 'm':
				args.promisc = 1;

				break;
				
			case 'j':
				if (args.jump == NULL) {
					args.jump = (char*) strdup(optarg);
				}

				break;

			case 'x':
				if (args.iptables == NULL) {
					args.iptables = (char*) strdup(optarg);
				}

				break;

			case 'y':
				if (args.ip6tables == NULL) {
					args.ip6tables = (char*) strdup(optarg);
				}

				break;

			case 'H':
#ifdef HAVE_MYSQL
				ptr = NULL;

				args.db_host = (char*) strndup(optarg, 132);

				for (i = 0; i < strlen(args.db_host); i++) {
					if (args.db_host[i] == ':') {
						args.db_host[i] = '\0';
						ptr = args.db_host + i + 1;
					}
				}

				if (ptr) {
					args.db_port = atoi(ptr);
				} else {
					args.db_port = DEF_DBPORT;
				}
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbhost' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;

			case 'K':
#ifdef HAVE_MYSQL
				args.db_sock = (char*) strndup(optarg, 255);
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbsock' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;

			case 'N':
#ifdef HAVE_MYSQL
				args.db_name = (char*) strndup(optarg, 132);
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbname' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;

			case 'U':
#ifdef HAVE_MYSQL
				args.db_user = (char*) strndup(optarg, 132);
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbuser' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;

			case 'W':
#ifdef HAVE_MYSQL
				args.db_pass = (char*) strndup(optarg, 132);
				memset(optarg, 'x', strlen(optarg) < 132 ? strlen(optarg) : 132);
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbpass' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;

			case 'M':
#ifdef HAVE_MYSQL
				args.db_comp = 1;
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbcomp' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;

			case 'G':
#ifdef HAVE_MYSQL
				for (i = 0; i < MAX_ALGO_NAMES; i++) {
					if (strncasecmp(optarg, algo_names[i], strlen(algo_names[i])) == 0) {
						args.db_algo = i;

						break;
					}
				}
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbalgo' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;
				
			case 'L':
#ifdef HAVE_MYSQL
				args.db_threshold = strtoul(optarg, NULL, 10);

				if (args.db_threshold < MIN_DBTHRS || args.db_threshold > MAX_DBTHRS) { /* Ignore invalid range */
					args.db_threshold = 0;
				}
#else
				if (warn == 0) {
					fprintf(stderr, "Support for '--dbthreshold' is not compiled into the binary.\n");
					warn = 1;
				}
#endif /* HAVE_MYSQL */
				break;
				
			case 'C':
				if (args.config == NULL) {
					args.config = (char*) strdup(optarg);
				}

				break;

			case 'u':
				if (args.user == NULL) {
					args.user = (char*) strdup(optarg);
				}

				break;

			case 'g':
				if (args.group == NULL) {
					args.group = (char*) strdup(optarg);
				}

				break;

			case 'r':
				if (args.chroot == NULL) {
					args.chroot = (char*) strdup(optarg);
				}

				break;

			case 'o':
				if (args.stats == NULL) {
					if (optarg) {
						args.stats = (char*) strdup(optarg);
					} else {
						args.stats = (char*) strdup(DEF_STATS);
					}
				}

				break;

			case 'z':
				if (args.dump == NULL) {
					if (optarg) {
						args.dump = (char*) strdup(optarg);
					} else {
						args.dump = (char*) strdup(DEF_DUMP);
					}
				}

				break;

			case 'n':
				if (optarg) {
					args.nice = atoi(optarg);
				}

				if (args.nice < -19 || args.nice > 20) {
					args.nice = DEF_NICENESS;
				}

				break;
				
			case 'd':
				if (optarg) {
					args.debug = strtoul(optarg, NULL, 10);

					if (args.debug > DEBUG_MAXVAL) {
						args.debug = DEBUG_MAXVAL;
					}
				} else {
					args.debug = DEBUG_ERROR;
				}

				break;

			case 'v':
				version(0);

				return EXIT_FAILURE;

			default:
			case 'h':
				usage();
				version(1);

				return EXIT_FAILURE;
		}
	}

	/* Exclude interface addresses */
	exclude_interfaces(&(args.exclude));

	if (optind < argc) {
		if (args.expression != NULL) {
			fprintf(stderr, "expression has been overridden by specified port and/or protocol\n");
		} else {
			args.expression = malloc(PCAP_EXPRESSION_SIZE);
			args.expression[0] = '\0';
			while (optind < argc) {
				if ((len = PCAP_EXPRESSION_SIZE - strlen(args.expression) - strlen(argv[optind])) > 0) {
					strncat(args.expression, argv[optind++], len);
					strncat(args.expression, " ", 1);
				} else {
					fprintf(stderr, "expression's size exceeds the internal buffer of (%u bytes)\n", PCAP_EXPRESSION_SIZE);

					return EXIT_FAILURE;
				}
			}
		}
	}

	/* Use behaviour of most commands (eg: SSH, Telnet). */
	if (args.ipv4 == USE_DEFAULT && args.ipv6 == 1) {
		args.ipv4 = 0;
	} else if (args.ipv6 == USE_DEFAULT && args.ipv4 == 1) {
		args.ipv6 = 0;
	}

	return EXIT_SUCCESS;
}

int process_config(struct config_node *node) {
	int i, valid, warn = 0;
	char buf[512], *ptr, *protocol = NULL, *port = NULL;
	struct threshold_t **t;
	struct prefix_t **p;
#ifdef HAVE_NET_SNMP
	char *snmpport = NULL;
	snmptrap_peer_t **s;
#endif
	struct in_addr ip;
	struct in6_addr ip6;
	struct config_node *n;

	while (node != NULL) {
		/* Interface */
		if (strcasecmp(node->var, "interface") == 0 && node->val != NULL) {
			if (args.interface == NULL) {
				args.interface = (char*) strdup(node->val);
			}

		/* IPv4 */
		} else if (strcasecmp(node->var, "ipv4") == 0) {
			if (args.ipv4 == USE_DEFAULT) {
				if (node->val == NULL) { /* Keyword only - enabled */
					args.ipv4 = 1;
				} else if (strncasecmp(node->val, "of", 2) == 0 || tolower(node->val[0]) == 'n' || node->val[0] == '0') { /* Off, No or 0 */
					args.ipv4 = 0;
				} else {
					args.ipv4 = 1;
				}
			}

		/* IPv6 */
		} else if (strcasecmp(node->var, "ipv6") == 0) {
			if (args.ipv6 == USE_DEFAULT) {
				if (node->val == NULL) { /* Keyword only - enabled */
					args.ipv6 = 1;
				} else if (strncasecmp(node->val, "of", 2) == 0 || tolower(node->val[0]) == 'n' || node->val[0] == '0') { /* Off, No or 0 */
					args.ipv6 = 0;
				} else {
					args.ipv6 = 1;
				}
			}

		/* Protocol - Used to build expression */
		} else if (strcasecmp(node->var, "protocol") == 0 && node->val != NULL) {
			if (protocol) free(protocol);
			protocol = strdup(node->val);

		/* Port - Used to build expression */
		} else if (strcasecmp(node->var, "port") == 0 && node->val != NULL) {
			if (port) free(port);
			port = strdup(node->val);

		/* Expression - tcpdump format */
		} else if (strcasecmp(node->var, "expression") == 0 && node->val != NULL) {
			if (args.expression) free(args.expression);
			args.expression = strdup(node->val);

		/* Threshold */
		} else if (strcasecmp(node->var, "threshold") == 0) {
			t = &(args.threshold);
			while ((*t) != NULL) t = &((*t)->next); /* Move to the end of threshold list */
			*t = malloc(sizeof(struct threshold_t));
			memset(*t, 0, sizeof(struct threshold_t));

			(*t)->rate = USE_DEFAULT;
			(*t)->burst = USE_DEFAULT;
			(*t)->decay = USE_DEFAULT;

			if (node->val != NULL) {
				(*t)->threshold = strtoul(node->val, NULL, 10);

				if ((*t)->threshold > MAX_THRESHOLD) {
					(*t)->threshold = MAX_THRESHOLD;
				}
			}

			n = node->sub;
			while (n != NULL) {
				if (strcasecmp(n->var, "pps") == 0 && n->val != NULL) {
					(*t)->threshold = strtoul(n->val, NULL, 10);

					if ((*t)->threshold > MAX_THRESHOLD) {
						(*t)->threshold = MAX_THRESHOLD;
					}
				} else if (strcasecmp(n->var, "rate") == 0 && n->val != NULL) {
					(*t)->rate = strtoul(n->val, NULL, 10);

					if ((*t)->rate > MAX_RATELIMIT) {
						(*t)->rate = MAX_RATELIMIT;
					}
				} else if (strcasecmp(n->var, "burst") == 0 && n->val != NULL) {
					(*t)->burst = strtoul(n->val, NULL, 10);

					if ((*t)->burst > MAX_BURST) {
						(*t)->burst = MAX_BURST;
					}
				} else if (strcasecmp(n->var, "decay") == 0 && n->val != NULL) {
					(*t)->decay = strtoul(n->val, NULL, 10);

					if ((*t)->decay > MAX_DECAY) {
						(*t)->decay = MAX_DECAY;
					}
				} else {
					syslog(LOG_INFO, "invalid option: <%s> %s", node->var, n->var);
				}

				n = n->next;
			}

			/* Remove threshold if not valid */
			if ((*t)->threshold == 0) {
				syslog(LOG_INFO, "invalid option: <%s> (missing 'pps')", node->var);
				free(*t);
				*t = NULL;
			}

		/* Rate - default value */
		} else if (strcasecmp(node->var, "rate") == 0 && node->val != NULL) {
			if (args.rate == DEF_RATELIMIT) {
				args.rate = strtoul(node->val, NULL, 10);

				if (args.rate > MAX_RATELIMIT) {
					args.rate = MAX_RATELIMIT;
				}
			}

		/* Burst - default value */
		} else if (strcasecmp(node->var, "burst") == 0 && node->val != NULL) {
			if (args.burst == DEF_BURST) {
				args.burst = strtoul(node->val, NULL, 10);

				if (args.burst > MAX_BURST) {
					args.burst = MAX_BURST;
				}
			}

		/* Decay - default value */
		} else if (strcasecmp(node->var, "decay") == 0 && node->val != NULL) {
			if (args.decay == DEF_DECAY) {
				args.decay = strtoul(node->val, NULL, 10);

				if (args.decay > MAX_DECAY) {
					args.decay = MAX_DECAY;
				}
			}

		/* Exclude - ignore traffic from exclusions */
		} else if (strcasecmp(node->var, "exclude") == 0) {
			p = &(args.exclude);
			while ((*p) != NULL) p = &((*p)->next); /* Move to the end of exclude list */

			n = node->sub;
			while (n != NULL) {
				*p = malloc(sizeof(struct prefix_t));
				memset(*p, 0, sizeof(struct prefix_t));
				valid = 0;

				if (strcasecmp(n->var, "prefix") == 0 && n->val != NULL) {
					if (strchr(n->val, ':') == NULL) {
						(*p)->ip_ss.ss_family = AF_INET;
						(*p)->mask_ss.ss_family = AF_INET;

						if ((ptr = strchr(n->val, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
							*ptr = '\0'; /* Separate mask from ip */
							ptr++; /* Move past seperator */

							(*p)->cidr = (strtoul(ptr, NULL, 10) < 32) ? strtoul(ptr, NULL, 10) : 32;
						} else {
							(*p)->cidr = 32;
						}

						cidr_to_mask_in_addr(&ip, (*p)->cidr);
						set_in_addr(&((*p)->mask_ss), &ip);

						if (inet_pton(AF_INET, n->val, &ip) == 0) {
							syslog(LOG_INFO, "excluded prefix '%s' has an invalid format\n", n->val);
						} else {
							set_in_addr(&((*p)->ip_ss), &ip);
							ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);

							valid = 1;
						}
					} else {
						/* IPv6 */
						(*p)->ip_ss.ss_family = AF_INET6;
						(*p)->mask_ss.ss_family = AF_INET6;

						if ((ptr = strchr(n->val, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
							*ptr = '\0'; /* Separate mask from ip */
							ptr++; /* Move past seperator */

							(*p)->cidr = (strtoul(ptr, NULL, 10) < 128) ? strtoul(ptr, NULL, 10) : 128;
						} else {
							(*p)->cidr = 128;
						}

						cidr_to_mask_in6_addr(&ip6, (*p)->cidr);
						set_in6_addr(&((*p)->mask_ss), &ip6);

						if (inet_pton(AF_INET6, n->val, &ip6) == 0) {
							syslog(LOG_INFO, "excluded prefix '%s' has an invalid format\n", n->val);
						} else {
							set_in6_addr(&((*p)->ip_ss), &ip6);
							ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);

							valid = 1;
						}
					}
				} else {
					syslog(LOG_INFO, "invalid option: <%s> %s", node->var, n->var);
				}

				if (valid == 1) { /* If valid, move to next prefix */
					p = &((*p)->next);
				} else { /* Remove invalid entry */
//					syslog(LOG_INFO, "invalid option: <%s> (missing 'prefix')", node->var);
					free(*p);
					*p = NULL;
				}

				/* If end if sub list, append block value prefix */
				if (n->next == NULL && node->val != NULL) {
					n->next = malloc(sizeof(struct config_node));
					memset(n->next, 0, sizeof(struct config_node));

					n->next->var = strdup("prefix");
					n->next->val = strdup(node->val);
					free(node->val);
					node->val = NULL;
				}

				n = n->next;
			}
		/* Include - include specified prefix rather than all traffic */
		} else if (strcasecmp(node->var, "include") == 0) {
			p = &(args.include);
			while ((*p) != NULL) p = &((*p)->next); /* Move to the end of include list */

			n = node->sub;
			while (n != NULL) {
				*p = malloc(sizeof(struct prefix_t));
				memset(*p, 0, sizeof(struct prefix_t));
				valid = 0;

				if (strcasecmp(n->var, "prefix") == 0 && n->val != NULL) {
					if (strchr(n->val, ':') == NULL) {
						(*p)->ip_ss.ss_family = AF_INET;
						(*p)->mask_ss.ss_family = AF_INET;

						if ((ptr = strchr(n->val, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
							*ptr = '\0'; /* Separate mask from ip */
							ptr++; /* Move past seperator */

							(*p)->cidr = (strtoul(ptr, NULL, 10) < 32) ? strtoul(ptr, NULL, 10) : 32;
						} else {
							(*p)->cidr = 32;
						}

						cidr_to_mask_in_addr(&ip, (*p)->cidr);
						set_in_addr(&((*p)->mask_ss), &ip);

						if (inet_pton(AF_INET, n->val, &ip) == 0) {
							syslog(LOG_INFO, "included prefix '%s' has an invalid format\n", n->val);
						} else {
							set_in_addr(&((*p)->ip_ss), &ip);
							ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);

							valid = 1;
						}
					} else {
						/* IPv6 */
						(*p)->ip_ss.ss_family = AF_INET6;
						(*p)->mask_ss.ss_family = AF_INET6;

						if ((ptr = strchr(n->val, '/')) != NULL && strlen(ptr) > 1) { /* Mask */
							*ptr = '\0'; /* Separate mask from ip */
							ptr++; /* Move past seperator */

							(*p)->cidr = (strtoul(ptr, NULL, 10) < 128) ? strtoul(ptr, NULL, 10) : 128;
						} else {
							(*p)->cidr = 128;
						}

						cidr_to_mask_in6_addr(&ip6, (*p)->cidr);
						set_in6_addr(&((*p)->mask_ss), &ip6);

						if (inet_pton(AF_INET6, n->val, &ip6) == 0) {
							syslog(LOG_INFO, "included prefix '%s' has an invalid format\n", n->val);
						} else {
							set_in6_addr(&((*p)->ip_ss), &ip6);
							ss_apply_cidr(&((*p)->ip_ss), (*p)->cidr);

							valid = 1;
						}
					}
				} else {
					syslog(LOG_INFO, "invalid option: <%s> %s", node->var, n->var);
				}

				if (valid == 1) { /* If valid, move to next prefix */
					p = &((*p)->next);
				} else { /* Remove invalid entry */
//					syslog(LOG_INFO, "invalid option: <%s> (missing 'prefix')", node->var);
					free(*p);
					*p = NULL;
				}

				/* If end if sub list, append block value prefix */
				if (n->next == NULL && node->val != NULL) {
					n->next = malloc(sizeof(struct config_node));
					memset(n->next, 0, sizeof(struct config_node));

					n->next->var = strdup("prefix");
					n->next->val = strdup(node->val);
					free(node->val);
					node->val = NULL;
				}

				n = n->next;
			}
		/* Algorithim used to caculate pps */
		} else if (strcasecmp(node->var, "algo") == 0 && node->val != NULL) {
			if (args.algo == DEF_PPS_ALGO) {
				for (i = 0; i < MAX_ALGO_NAMES; i++) {
					if (strncasecmp(node->val, algo_names[i], strlen(algo_names[i])) == 0) {
						args.algo = i;

						break;
					}
				}
			}

		/* Interval - Length of time to calculate PPS over. */
		} else if (strcasecmp(node->var, "interval") == 0 && node->val != NULL) {
			if (args.interval == DEF_INTERVAL) {
				args.interval = strtoul(node->val, NULL, 10);

				if (args.interval > MAX_INTERVAL) {
					args.interval = MAX_INTERVAL;
				} else if (args.interval < MIN_INTERVAL) {
					args.interval = MIN_INTERVAL;
				}

				args.stale_interval = (args.interval * STALE_INTERVAL_MLT);
				args.max_samples = (args.interval + 1);
			}

		} else if (strcasecmp(node->var, "calc-interval") == 0 && node->val != NULL) {
			if (args.check_interval == DEF_CHECK_INTERVAL) {
				args.check_interval = strtoul(node->val, NULL, 10);

				if (args.check_interval > MAX_CHECK_INTERVAL) {
					args.check_interval = MAX_CHECK_INTERVAL;
				} else if (args.check_interval < MIN_CHECK_INTERVAL) {
					args.check_interval = MIN_CHECK_INTERVAL;
				}
			}

		} else if (strcasecmp(node->var, "stats-interval") == 0 && node->val != NULL) {
			if (args.stats_interval == DEF_STATS_INTERVAL) {
				args.stats_interval = strtoul(node->val, NULL, 10);

				if (args.stats_interval > MAX_STATS_INTERVAL) {
					args.stats_interval = MAX_STATS_INTERVAL;
				} else if (args.stats_interval < MIN_STATS_INTERVAL) {
					args.stats_interval = MIN_STATS_INTERVAL;
				}
			}

		} else if (strcasecmp(node->var, "snmptrap") == 0) {
#ifdef HAVE_NET_SNMP
			s = &(args.peer);
			while ((*s) != NULL) s = &((*s)->next); /* Move to the end of peer list */
			*s = malloc(sizeof(snmptrap_peer_t));
			memset(*s, 0, sizeof(snmptrap_peer_t));

			strncpy((*s)->community, "public", sizeof((*s)->community));
			(*s)->version = SNMP_VERSION_2c;

			/* Capture peername (if available) */
			if (node->val != NULL) {
				strncpy((*s)->peername, node->val, sizeof((*s)->peername));
			}

			n = node->sub;
			while (n != NULL) {
				if (strcasecmp(n->var, "host") == 0 && n->val != NULL) {
					strncpy((*s)->peername, n->val, sizeof((*s)->peername));
				} else if (strcasecmp(n->var, "port") == 0 && n->val != NULL) {
					if (atoi(n->val) >= 1 && atoi(n->val) <= 65535) { /* Ignore invalid range */
						if (snmpport) free(snmpport);
						snmpport = strdup(n->val);
					}
				} else if (strcasecmp(n->var, "community") == 0 && n->val != NULL) {
					strncpy((*s)->community, n->val, sizeof((*s)->community));
				} else if (strcasecmp(n->var, "version") == 0 && n->val != NULL) {
					if (strcasecmp(n->val, "1") == 0 || strcasecmp(n->val, "v1") == 0) {
						(*s)->version = SNMP_VERSION_1;
					} else if (strcasecmp(n->val, "3") == 0 || strcasecmp(n->val, "v3") == 0) {
						(*s)->version = SNMP_VERSION_3;
					} else {
						(*s)->version = SNMP_VERSION_2c;
					}
				} else if (strcasecmp(n->var, "inform") == 0) {
					if (n->val == NULL) { /* Keyword only - enabled */
						(*s)->inform = 1;
					} else if (strncasecmp(n->val, "y", 1) == 0 || strcasecmp(n->val, "on") == 0 || strcasecmp(n->val, "1") == 0) { /* Yes, On or 1 */
						(*s)->inform = 1;
					} else {
						(*s)->inform = 0;
					}
				} else {
					syslog(LOG_INFO, "invalid option: <%s> %s", node->var, n->var);
				}

				n = n->next;
			}

			/* Remove peer if not valid */
			if ((*s)->peername[0] == '\0') {
				syslog(LOG_INFO, "invalid option: <%s> (missing 'host')", node->var);
				free(*s);
				*s = NULL;
			} else if (snmpport != NULL && strchr((*s)->peername, ':') == NULL) {
				snprintf(buf, sizeof(buf), "%s:%u", (*s)->peername, atoi(snmpport));
				strncpy((*s)->peername, buf, sizeof((*s)->peername));
			}

			if (snmpport) {
				free(snmpport);
				snmpport = NULL;
			}
#else
			if (warn == 0) {
				syslog(LOG_INFO, "Support for SNMPTrap is not compiled into the binary.");
				warn = 1;
			}
#endif /* HAVE_NET_SNMP */
		/* Test mode */
		} else if (strcasecmp(node->var, "test") == 0) {
			args.test = 1;
		/* Direction of Capture */
		} else if (strcasecmp(node->var, "direction") == 0 && node->val != NULL) {
			if (strncasecmp(node->var, "inout", 5) == 0) {
				args.direction = PCAP_D_INOUT;
			} else if (strncasecmp(node->var, "out", 3) == 0) {
				args.direction = PCAP_D_OUT;
			} else if (strncasecmp(node->var, "in", 2) == 0) {
				args.direction = PCAP_D_IN;
			}
		/* Promiscuous mode */
		} else if (strcasecmp(node->var, "promiscuous") == 0) {
			args.promisc = 1;

		} else if (strcasecmp(node->var, "jump") == 0 && node->val != NULL) {
			if (args.jump == NULL) {
				args.jump = strdup(node->val);
			}

		} else if (strcasecmp(node->var, "iptables") == 0 && node->val != NULL) {
			if (args.iptables == NULL) {
				args.iptables = strdup(node->val);
			}

		} else if (strcasecmp(node->var, "ip6tables") == 0 && node->val != NULL) {
			if (args.ip6tables == NULL) {
				args.ip6tables = strdup(node->val);
			}

#ifdef HAVE_MYSQL
		} else if (strcasecmp(node->var, "database") == 0) {
			if (args.db_host == NULL) {
				/* Capture database host (if available) */
				if (node->val != NULL) {
					args.db_host = (char*) strndup(node->val, 132);
				}

				n = node->sub;
				while (n != NULL) {
					if (strncasecmp(n->var, "host", 4) == 0 && n->val != NULL) {
						if (args.db_host == NULL) {
							args.db_host = (char*) strndup(n->val, 132);
						}

					} else if (strcasecmp(n->var, "port") == 0 && n->val != NULL) {
						if (args.db_port == 0 && atoi(n->val) >= 1 && atoi(n->val) <= 65535) { /* Ignore invalid range */
							args.db_port = atoi(n->val);
						}

					} else if (strncasecmp(n->var, "sock", 4) == 0 && n->val != NULL) {
						if (args.db_sock == NULL) {
							args.db_sock = (char*) strndup(n->val, 255);
						}

					} else if (strcasecmp(n->var, "name") == 0 && n->val != NULL) {
						if (args.db_name == NULL) {
							args.db_name = (char*) strndup(n->val, 132);
						}

					} else if (strncasecmp(n->var, "user", 4) == 0 && n->val != NULL) {
						if (args.db_user == NULL) {
							args.db_user = (char*) strndup(n->val, 132);
						}

					} else if (strncasecmp(n->var, "pass", 4) == 0 && n->val != NULL) {
						if (args.db_pass == NULL) {
							args.db_pass = (char*) strndup(n->val, 132);
						}

					} else if (strncasecmp(n->var, "comp", 4) == 0) {
						if (args.db_comp == 0) {
							if (n->val == NULL || strcasecmp(n->var, "yes") == 0 || strcasecmp(n->var, "on") == 0 || strcasecmp(n->var, "1") == 0) {
								args.db_comp = 1;
							}
						}

					/* Algorithim used to caculate pps */
					} else if (strncasecmp(node->var, "algo", 4) == 0 && n->val != NULL) {
						if (args.db_algo == DEF_DBALGO) {
							for (i = 0; i < MAX_ALGO_NAMES; i++) {
								if (strncasecmp(n->val, algo_names[i], strlen(algo_names[i])) == 0) {
									args.db_algo = i;

									break;
								}
							}
						}

					} else if (strcasecmp(n->var, "threshold") == 0 && n->val != NULL) {
						if (args.db_threshold == 0 && atoi(n->val) >= MIN_DBTHRS && atoi(n->val) <= MAX_DBTHRS) { /* Ignore invalid range */
							args.db_threshold = atoi(n->val);
						}

					} else {
						syslog(LOG_INFO, "invalid option: <%s> %s", node->var, n->var);
					}

					n = n->next;
				}
			}
#else
			if (warn == 0) {
				syslog(LOG_INFO, "Support for MySQL is not compiled into the binary.");
				warn = 1;
			}
#endif /* HAVE_MYSQL */

		} else if (strcasecmp(node->var, "user") == 0 && node->val != NULL) {
			if (args.user == NULL) {
				args.user = strdup(node->val);
			}

		} else if (strcasecmp(node->var, "group") == 0 && node->val != NULL) {
			if (args.group == NULL) {
				args.group = strdup(node->val);
			}

		} else if (strcasecmp(node->var, "chroot") == 0 && node->val != NULL) {
			if (args.chroot == NULL) {
				args.chroot = strdup(node->val);
			}

		} else if (strcasecmp(node->var, "stats") == 0) {
			if (args.stats == NULL) {
				if (node->val) {
					args.stats = (char*) strdup(node->val);
				} else {
					args.stats = (char*) strdup(DEF_STATS);
				}
			}

		} else if (strcasecmp(node->var, "dump") == 0) {
			if (args.dump == NULL) {
				if (node->val) {
					args.dump = (char*) strdup(node->val);
				} else {
					args.dump = (char*) strdup(DEF_DUMP);
				}
			}

		} else if (strcasecmp(node->var, "nice") == 0 && node->val != NULL) {
			if (args.nice == DEF_NICENESS) {
				args.nice = atoi(node->val);

				if (args.nice < -19 || args.nice > 20) {
					args.nice = DEF_NICENESS;
				}
			}

		} else if (strcasecmp(node->var, "debug") == 0) {
			if (args.debug == 0) {
				if (node->val != NULL) {
					args.debug = strtoul(node->val, NULL, 10);

					if (args.debug > DEBUG_MAXVAL) {
						args.debug = DEBUG_MAXVAL;
					}
				} else {
					args.debug = DEBUG_ERROR;
				}
			}

		} else {
			syslog(LOG_INFO, "invalid option: %s", node->var);
		}

		node = node->next;
	}

	/* Build expression from protocol and port if not set. */
	if (args.expression == NULL) {
		if (protocol != NULL && port != NULL) {
			snprintf(buf, sizeof(buf), "%s and dst port %s", protocol, port);
			args.expression = malloc(strlen(buf) + 1);
			strncpy(args.expression, buf, strlen(buf));
		} else if (protocol != NULL) {
			args.expression = strdup(protocol);
		} else if (port != NULL) {
			snprintf(buf, sizeof(buf), "dst port %s", port);
			args.expression = malloc(strlen(buf) + 1);
			strncpy(args.expression, buf, strlen(buf));
		}
	}

	/* Cleanup */
	if (protocol) free(protocol);
	if (port) free(port);

	return 0;
}

int set_defaults() {
	if (args.db_name == NULL) {
		args.db_name = strdup(DEF_DBNAME);
	}

	if (args.db_user == NULL) {
		args.db_user = strdup(DEF_DBUSER);
	}

	/* Default to both IPv4 and IPv6 */
	if (args.ipv4 == USE_DEFAULT && args.ipv6 == USE_DEFAULT) {
		args.ipv4 = DEF_IPV4;
		args.ipv6 = DEF_IPV6;
	}

	if (args.jump == NULL) {
		args.jump = (char*) strdup(DEF_JUMP);
	}

	if (args.iptables == NULL) {
		args.iptables = (char*) strdup(DEF_IPTABLES);
	}

	if (args.ip6tables == NULL) {
		args.ip6tables = (char*) strdup(DEF_IP6TABLES);
	}

	return 0;
}

int exclude_interfaces(struct prefix_t **exclude) {
	char errbuf[PCAP_ERRBUF_SIZE];
	char buf[46];
	pcap_if_t* pcap_if;
	pcap_if_t *alldevs = NULL;
	pcap_addr_t* pcap_addr;
	struct prefix_t **p;
	struct in_addr in;
	struct in6_addr in6;

	union __sockaddr_union {
		struct sockaddr_storage *ss;
		struct sockaddr *sa;
		struct sockaddr_in *sa_in;
		struct sockaddr_in6 *sa_in6;
	} u;

	if (pcap_findalldevs(&alldevs, errbuf) == -1) {
		fprintf(stderr, "pcap_findalldevs(): %s\n", errbuf);

		return 1;
	}

	p = exclude;
	while ((*p) != NULL) p = &((*p)->next); /* Move to the end of exclude list */

	pcap_if = alldevs;
	while (pcap_if != NULL) {
		pcap_addr = pcap_if->addresses;
		while (pcap_addr != NULL) {
			if (pcap_addr->addr != NULL) {
				u.sa = (struct sockaddr*) pcap_addr->addr;

				if ((u.sa->sa_family == AF_INET || u.sa->sa_family == AF_INET6) && pcap_addr->addr != NULL && pcap_addr->netmask != NULL) {
					if (args.debug >= DEBUG_INTERNAL) {
						inet_ntop(u.sa->sa_family, get_in_addr2(u.sa), buf, sizeof(buf)); /* REMOVE */
						//fprintf(stderr, "Excluding %s/%u (interface %s)\n", buf, (u.sa->sa_family == AF_INET) ? 32 : 128, pcap_if->name);
					}

					*p = malloc(sizeof(struct prefix_t));
					memset(*p, 0, sizeof(struct prefix_t));

					if (u.sa->sa_family == AF_INET) {
						(*p)->cidr = 32;
						set_in_addr(&((*p)->ip_ss), &(u.sa_in->sin_addr));
						cidr_to_mask_in_addr(&in, (*p)->cidr);
						set_in_addr(&((*p)->mask_ss), &in);
					} else {
						(*p)->cidr = 128;
						set_in6_addr(&((*p)->ip_ss), &(u.sa_in6->sin6_addr));
						cidr_to_mask_in6_addr(&in6, (*p)->cidr);
						set_in6_addr(&((*p)->mask_ss), &in6);
					}
					(*p)->interface = 1;

					/* Update end of list */
					p = &((*p)->next);
				}
			}
			pcap_addr = pcap_addr->next;
		}
		pcap_if = pcap_if->next;
	}

	if (alldevs) {
		pcap_freealldevs(alldevs);
	}

	return 0;
}

void usage(void) {
	printf("Usage: %s [OPTION] <-T PPS[:RATE[:BURST[:DECAY]]] ...> [expression]\n", PROGRAM_NAME);
	printf("Per Host RatE Limiter (%s) - Generates and updates iptables chains for hosts\nwhich cross the configured threshold(s).\n", PROGRAM_PKGN);
	printf("\nPreferences:\n");
	printf("  -I, --interface=INTERFACE           monitor traffic on interface (any)\n");
	printf("  -4, --ipv4                          monitor IPv4 traffic (on)\n");
	printf("  -6, --ipv6                          monitor IPv6 traffic (on)\n");
	printf("  -P, --protocol=PROTO                monitor traffic for protocol (any)\n");
	printf("  -p, --port=PORT                     monitor traffic on port (any)\n");
	printf("  -T, --threshold=PPS[:R[:B[:D]]]     user threshold (-:%d:%d:%d)\n", DEF_RATELIMIT, DEF_BURST, DEF_DECAY);
	printf("  -R, --rate=PPS                      use rate limit (%d)\n", DEF_RATELIMIT);
	printf("  -B, --burst=PACKETS                 use burst for rate limiter (%d)\n", DEF_BURST);
	printf("  -D, --decay=SECONDS                 use decay (%d)\n", DEF_DECAY);
	printf("  -X, --exclude=PREFIX[/CIDR]         exclude packets from prefix (none)\n");
	printf("  -Y, --include=PREFIX[/CIDR]         include packets from prefix (any)\n");
	printf("  -A, --algo=<avg|sum|max>            use algorithim to calculate PPS (%s)\n", algo_names[DEF_PPS_ALGO]);
	printf("  -i, --interval=SEC                  interval to calculate PPS over (%d)\n", DEF_INTERVAL);
	printf("  -c, --check-interval=SEC            interval to check thresholds (%d)\n", DEF_CHECK_INTERVAL);
	printf("  -s, --stats-interval=SEC            interval to update stats file (%d)\n", DEF_STATS_INTERVAL);
	printf("  -S, --snmptrap=HOST[:P][:C[:V[:I]]] send snmptraps to host (none)\n");
	printf("  -t, --test                          test mode - do not update iptables (off)\n");
	printf("  -w, --direction=<in|out|inout>      direction to capture packets (in)\n");
	printf("  -m, --promiscuous                   promiscuous mode (off)\n");
	printf("  -j, --jump=CHAIN                    jump to user defined chain (%s)\n", DEF_JUMP);
	printf("  -x, --iptables=PATH                 full path to iptables (%s)\n", DEF_IPTABLES);
	printf("  -y, --ip6tables=PATH                full path to ip6tables (%s)\n", DEF_IP6TABLES);
	printf("\nDatabase:\n");
	printf("  -H, --dbhost=HOST[:PORT]            MySQL host for data syncronization (%s)\n", strlen(DEF_DBHOST) > 1 ? DEF_DBHOST : "none");
	printf("  -K, --dbsock=SOCKET                 MySQL socket for data syncronization (%s)\n", strlen(DEF_DBSOCK) > 1 ? DEF_DBSOCK : "none");
	printf("  -N, --dbname=DATABASE               MySQL database to store data within (%s)\n", strlen(DEF_DBNAME) > 1 ? DEF_DBNAME : "none");
	printf("  -U, --dbuser=USER                   use specified username for connecting to the database (%s)\n", strlen(DEF_DBUSER) > 1 ? DEF_DBUSER : "none");
	printf("  -W, --dbpass=PASSWORD               use specified password for connecting to the database (%s)\n", strlen(DEF_DBPASS) > 1 ? DEF_DBPASS : "none");
	printf("  -M, --dbcomp                        use compression protocol to communicate with the database (off)\n");
	printf("  -G, --dbalgo=<avg|sum|max>          use algorithim to calculate PPS with database input (%s)\n", algo_names[DEF_DBALGO]);
	printf("  -L, --dbthreshold=PPS               threshold for data to be written to database (%u)\n", DEF_DBTHRS);
	printf("\nProcess:\n");
	printf("  -C, --config=FILE                   use FILE for configuration (none)\n");
	printf("  -u, --user=USER                     run process as USER (current)\n");
	printf("  -g, --group=GROUP                   run process as GROUP (current)\n");
	printf("  -r, --chroot=DIR                    change root to DIR (none)\n");
	printf("  -o, --stats[=FILE]                  write stats to FILE (%s)\n", strlen(DEF_STATS) > 1 ? DEF_STATS : "none");
	printf("  -z, --dump[=FILE]                   write dump to FILE (%s)\n", strlen(DEF_DUMP) > 1 ? DEF_DUMP : "none");
	printf("  -n, --nice=[+/-]NUM                 set process niceness (%s%d)\n", (DEF_NICENESS) >= 0 ? "+" : "-", DEF_NICENESS);
	printf("  -d, --debug[=LEVEL]                 enable debug and remain in foreground (off)\n");
	printf("\nGeneral:\n");
	printf("  -v, --version                       displays version information\n");
	printf("  -h, --help                          displays this help information\n\n");
	printf("Examples:\n\n");
	printf("  %s -p 53 -T 50:5 -T 75\n", PROGRAM_NAME);
	printf("  %s -p 53 -T 25:20 -T 50:10 -T 75 --snmptrap=192.168.33.1:public:v2\n", PROGRAM_NAME);
	printf("  %s -I eth0 -P tcp -p 25 -T 25:20:10:300 -T 50:10:5 -T 75:0\n", PROGRAM_NAME);
	printf("  %s -I eth0 -p 53 -T 100 -B 10 -D 1800 -X 192.168.33.1 -X 192.168.55.0/24\n", PROGRAM_NAME);
	printf("  %s -T 50 'udp dst port 53 and not (dst net 192.168.33.1 or 192.168.55.0/24)'\n", PROGRAM_NAME);
	printf("\n");
}

void version(int flag) {
	if (flag == 0 || flag == 2) {
		printf("%s v%s\n", PROGRAM_NAME, PROGRAM_VERS);
	}
	printf("Written by James M. Sella <%s>\n", PROGRAM_BUGS);
	if (!flag) {
		printf("\nCopyright (c) 2005-2013 James M. Sella. All rights reserved.\n");
		printf("This is free software; see the source for copying conditions. There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n");
	}
}

void display_args(void) {
	char buf[46], buf2[46];
	struct threshold_t *t;
	struct prefix_t *p;
	snmptrap_peer_t *s;

	syslog(LOG_DEBUG, "[Arguments]");

	if (args.interface) {
		syslog(LOG_DEBUG, " interface: %s", args.interface);
	} else {
		syslog(LOG_DEBUG, " interface: any");
	}

	syslog(LOG_DEBUG, " ipv4: %s", (args.ipv4) ? "on" : "off");
	syslog(LOG_DEBUG, " ipv6: %s", (args.ipv6) ? "on" : "off");

	if ((t = args.threshold) != NULL) {
		while (t != NULL) {
			syslog(LOG_DEBUG, " threshold: %u:%u:%u:%u (threshold:rate:burst:decay)", t->threshold, (t->rate == USE_DEFAULT) ? args.rate : t->rate, (t->burst == USE_DEFAULT) ? args.burst : t->burst, (t->decay == USE_DEFAULT) ? args.decay : t->decay);
			t = t->next;
		}
	} else {
		syslog(LOG_DEBUG, " threshold: none");
	}

	syslog(LOG_DEBUG, " rate: %u", args.rate);
	syslog(LOG_DEBUG, " burst: %u", args.burst);
	syslog(LOG_DEBUG, " decay: %u", args.decay);
	
	if ((p = args.exclude) != NULL) {
		while (p != NULL) {
			inet_ntop(p->ip_ss.ss_family, get_in_addr(&(p->ip_ss)), buf, sizeof(buf));
			inet_ntop(p->mask_ss.ss_family, get_in_addr(&(p->mask_ss)), buf2, sizeof(buf2));
			syslog(LOG_DEBUG, " exclude: %s/%u%s (/%u == %s)", buf, p->cidr, (p->interface) ? "*" : "", p->cidr, buf2);
			p = p->next;
		}
	} else {
		syslog(LOG_DEBUG, " exclude: none");
	}

	if ((p = args.include) != NULL) {
		while (p != NULL) {
			inet_ntop(p->ip_ss.ss_family, get_in_addr(&(p->ip_ss)), buf, sizeof(buf));
			inet_ntop(p->mask_ss.ss_family, get_in_addr(&(p->mask_ss)), buf2, sizeof(buf2));
			syslog(LOG_DEBUG, " include: %s/%u (/%u == %s)", buf, p->cidr, p->cidr, buf2);
			p = p->next;
		}
	} else {
		syslog(LOG_DEBUG, " include: 0.0.0.0/0 (any)");
	}

	syslog(LOG_DEBUG, " algo: %s", algo_names[args.algo]);

	syslog(LOG_DEBUG, " interval: %d", args.interval);
	syslog(LOG_DEBUG, " check interval: %d", args.check_interval);
	syslog(LOG_DEBUG, " stats interval: %d", args.stats_interval);

	if ((s = args.peer) != NULL) {
		while (s != NULL) {
			syslog(LOG_DEBUG, " snmptrap: %s:%s:%s:%d (host[:port]:community:version:inform)", s->peername, s->community, (s->version == SNMP_VERSION_1) ? "v1" : (s->version == SNMP_VERSION_3) ? "v3" : "v2c", s->inform);
			s = s->next;
		}
	} else {
		syslog(LOG_DEBUG, " snmptrap: none");
	}

	if (args.expression) {
		syslog(LOG_DEBUG, " expression: %s", args.expression);
	} else {
		syslog(LOG_DEBUG, " expression: none");
	}

	if (args.test > 0) {
		syslog(LOG_DEBUG, " test: enabled");
	} else {
		syslog(LOG_DEBUG, " test: disabled");
	}

	if (args.direction == PCAP_D_IN) {
		syslog(LOG_DEBUG, " direction: in");
	} else if (args.direction == PCAP_D_OUT) {
		syslog(LOG_DEBUG, " direction: out");
	} else {
		syslog(LOG_DEBUG, " direction: inout");
	}

	if (args.promisc > 0) {
		syslog(LOG_DEBUG, " promiscuous: enabled");
	} else {
		syslog(LOG_DEBUG, " promiscuous: disabled");
	}

	if (args.jump) {
		syslog(LOG_DEBUG, " jump: %s", args.jump);
	} else {
		syslog(LOG_DEBUG, " jump: none");
	}

	if (args.iptables) {
		syslog(LOG_DEBUG, " iptables: %s", args.iptables);
	} else {
		syslog(LOG_DEBUG, " iptables: none");
	}

	if (args.ip6tables) {
		syslog(LOG_DEBUG, " ip6tables: %s", args.ip6tables);
	} else {
		syslog(LOG_DEBUG, " ip6tables: none");
	}

	if (args.db_host) {
		syslog(LOG_DEBUG, " dbhost: %s", args.db_host);
	} else {
		syslog(LOG_DEBUG, " dbhost: none");
	}

	if (args.db_sock) {
		syslog(LOG_DEBUG, " dbsock: %s", args.db_sock);
	} else {
		syslog(LOG_DEBUG, " dbsock: none");
	}

	if (args.db_name) {
		syslog(LOG_DEBUG, " dbname: %s", args.db_name);
	} else {
		syslog(LOG_DEBUG, " dbname: none");
	}

	if (args.db_user) {
		syslog(LOG_DEBUG, " dbuser: %s", args.db_user);
	} else {
		syslog(LOG_DEBUG, " dbuser: none");
	}

	if (args.db_pass) {
		syslog(LOG_DEBUG, " dbpass: %s", args.db_pass);
	} else {
		syslog(LOG_DEBUG, " dbpass: none");
	}

	if (args.db_comp) {
		syslog(LOG_DEBUG, " dbcomp: %d", args.db_comp);
	} else {
		syslog(LOG_DEBUG, " dbcomp: none");
	}

	if (args.db_algo) {
		syslog(LOG_DEBUG, " dbalgo: %s", algo_names[args.db_algo]);
	} else {
		syslog(LOG_DEBUG, " dbalgo: none");
	}

	if (args.db_threshold) {
		syslog(LOG_DEBUG, " dbthreshold: %d", args.db_threshold);
	} else {
		syslog(LOG_DEBUG, " dbthreshold: none");
	}

	if (args.config) {
		syslog(LOG_DEBUG, " config: %s", args.config);
	} else {
		syslog(LOG_DEBUG, " config: none");
	}

	if (args.user) {
		syslog(LOG_DEBUG, " user: %s", args.user);
	} else {
		syslog(LOG_DEBUG, " user: current");
	}

	if (args.group) {
		syslog(LOG_DEBUG, " group: %s", args.group);
	} else {
		syslog(LOG_DEBUG, " group: current");
	}

	if (args.stats) {
		syslog(LOG_DEBUG, " stats: %s", args.stats);
	} else {
		syslog(LOG_DEBUG, " stats: disabled");
	}

	if (args.dump) {
		syslog(LOG_DEBUG, " dump: %s", args.dump);
	} else {
		syslog(LOG_DEBUG, " dump: disabled");
	}

	if (args.chroot) {
		syslog(LOG_DEBUG, " chroot: %s", args.chroot);
	} else {
		syslog(LOG_DEBUG, " chroot: disabled");
	}

	syslog(LOG_DEBUG, " nice: %d", args.nice);

	if (args.debug > 0) {
		syslog(LOG_DEBUG, " debug: %d", args.debug);
	} else {
		syslog(LOG_DEBUG, " debug: disabled");
	}
}

void free_args(void) {
	struct threshold_t *t;
	struct prefix_t *p;
	snmptrap_peer_t *s;

	if (args.interface) free(args.interface);
	if (args.expression) free(args.expression);
	if (args.jump) free(args.jump);
	if (args.iptables) free(args.iptables);
	if (args.ip6tables) free(args.ip6tables);
	if (args.db_host) free(args.db_host);
	if (args.db_sock) free(args.db_sock);
	if (args.db_name) free(args.db_name);
	if (args.db_user) free(args.db_user);
	if (args.db_pass) free(args.db_pass);
	if (args.config) free(args.config);
	if (args.user) free(args.user);
	if (args.group) free(args.group);
	if (args.chroot) free(args.chroot);
	if (args.stats) free(args.stats);
	while (args.threshold) {
		t = args.threshold;
		args.threshold = t->next;
		free(t);
	}
	while (args.exclude) {
		p = args.exclude;
		args.exclude = p->next;
		free(p);
	}
	while (args.include) {
		p = args.include;
		args.include = p->next;
		free(p);
	}
	while (args.peer) {
		s = args.peer;
		args.peer = s->next;
		free(s);
	}

	memset(&args, 0, sizeof(args));
}

/*
** Local Variables:
** c-basic-offset: 3
** tab-width: 3
** End:
** vim: noet ts=3 sw=3
*/
