/*
** PHREL
** $Id: util.c,v 1.16 2013/04/21 02:23:08 sella Exp $
** Copyright (c) 2004-2013 James M. Sella. All Rights Reserved.
** Released under the GPL Version 2 License.
** http://www.digitalgenesis.com
*/

static const char rcsid[] __attribute__((used)) = "$Id: util.c,v 1.16 2013/04/21 02:23:08 sella Exp $";

#include "util.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <assert.h>

union __sockaddr_union {
	struct sockaddr_storage *ss;
	struct sockaddr *sa;
	struct sockaddr_in *sa_in;
	struct sockaddr_in6 *sa_in6;
};

void *get_in_addr(const struct sockaddr_storage *ss) {
	union __sockaddr_union u;

	u.ss = (struct sockaddr_storage*) ss;
	if (ss->ss_family == AF_INET) {
		return &(u.sa_in->sin_addr);
	} else {
		return &(u.sa_in6->sin6_addr);
	}
}

void *get_in_addr2(const struct sockaddr *sa) {
	union __sockaddr_union u;

	u.sa = (struct sockaddr*) sa;
	if (sa->sa_family == AF_INET) {
		return &(u.sa_in->sin_addr);
	} else {
		return &(u.sa_in6->sin6_addr);
	}
}

void set_in_addr(struct sockaddr_storage *ss, struct in_addr *in) {
	union __sockaddr_union u;

	memset(ss, 0, sizeof(struct sockaddr_storage));

	u.ss = ss;
	u.sa_in->sin_family = AF_INET;
	memcpy(&(u.sa_in->sin_addr), in, sizeof(struct in_addr));
}

void set_in6_addr(struct sockaddr_storage *ss, struct in6_addr *in6) {
	union __sockaddr_union u;

	memset(ss, 0, sizeof(struct sockaddr_storage));

	u.ss = ss;
	u.sa_in6->sin6_family = AF_INET6;
	memcpy(&(u.sa_in6->sin6_addr), in6, sizeof(struct in6_addr));
}

/* FIXME: Recode using rtnetlink to get IPv6 also */
// Stolen from Newsgroup example by Floyd Davidson (floyd@ptialaska.net) on 2002-07-16 04:50:07 PST
void source_ip(char *interface, struct in_addr *ip) {
	int                sockfd, size  = 1;
	struct ifreq       *ifr;
	struct ifconf      ifc;
	struct sockaddr_in sa;

	if (0 > (sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP))) {
		syslog(LOG_EMERG, "Failed to open socket.");
		exit(EXIT_FAILURE);
	}

	ifc.ifc_len = IFRSIZE;
	ifc.ifc_req = NULL;

	do {
		++size;
		//FIXME: Does this realloc cause a leak??
		/* realloc buffer size until no overflow occurs  */
		if (NULL == (ifc.ifc_req = (struct ifreq*) realloc(ifc.ifc_req, IFRSIZE))) {
			syslog(LOG_EMERG, "Out of memory.");
			exit(EXIT_FAILURE);
		}
		ifc.ifc_len = IFRSIZE;
		if (ioctl(sockfd, SIOCGIFCONF, &ifc)) {
			syslog(LOG_EMERG, "ioctl(SIOCFIFCONF) error");
			exit(EXIT_FAILURE);
		}
	} while (IFRSIZE <= ifc.ifc_len);

	ifr = ifc.ifc_req;
	for (;(char *) ifr < (char *) ifc.ifc_req + ifc.ifc_len; ++ifr) {

		if (ifr->ifr_addr.sa_data == (ifr+1)->ifr_addr.sa_data) {
			continue;  /* duplicate, skip it */
		}

		if (ioctl(sockfd, SIOCGIFFLAGS, ifr)) {
			continue;  /* failed to get flags, skip it */
		}

		if (strncmp(ifr->ifr_name, "lo", 2) && (strlen(interface) < 1 || !strcmp(ifr->ifr_name, interface))) {
			//fprintf(stderr, "Interface: %s\n", ifr->ifr_name);
			*ip = inaddrr(ifr_addr.sa_data);
			//fprintf(stderr, "IP Address: %s\n", inet_ntoa(*ip));

			break;
		}
	}

	/* Free memory. */
	free(ifc.ifc_req);

	close(sockfd);
}

int ss_cmp(const struct sockaddr_storage *ip1, const struct sockaddr_storage *ip2, const unsigned int cidr) {
	struct in_addr in_ip1, in_ip2;
	struct in6_addr in6_ip1, in6_ip2;

	if (ip1->ss_family != ip2->ss_family) {
		return 1;
	}

	if ((ip1->ss_family == AF_INET && cidr > 32) || (ip1->ss_family == AF_INET6 && cidr > 128)) {
		return -1;
	}

	if (ip1->ss_family == AF_INET) {
		memcpy(&in_ip1, get_in_addr(ip1), sizeof(struct in_addr));
		memcpy(&in_ip2, get_in_addr(ip2), sizeof(struct in_addr));

		if (cidr < 32) {
			in_apply_cidr(&in_ip1, cidr);
			in_apply_cidr(&in_ip2, cidr);
		}

		return (memcmp(&in_ip1, &in_ip2, sizeof(struct in_addr))) ? 1 : 0;
	} else if (ip1->ss_family == AF_INET6) {
		memcpy(&in6_ip1, get_in_addr(ip1), sizeof(struct in6_addr));
		memcpy(&in6_ip2, get_in_addr(ip2), sizeof(struct in6_addr));

		if (cidr < 128) {
			in6_apply_cidr(&in6_ip1, cidr);
			in6_apply_cidr(&in6_ip2, cidr);
		}

		return (memcmp(&in6_ip1, &in6_ip2, sizeof(struct in6_addr))) ? 1 : 0;
	}

	return 1;
}

struct in_addr* cidr_to_mask_in_addr(struct in_addr* in, const unsigned int cidr) {
	if (cidr > 32) {
		return NULL;
	}

	in->s_addr = htonl(INADDR_BROADCAST);
	in->s_addr = htonl(ntohl(in->s_addr) >> (32 - cidr) << (32 - cidr));

	return in;
}

struct in6_addr* cidr_to_mask_in6_addr(struct in6_addr* in6, const unsigned int cidr) {
	int i;

	if (cidr > 128) {
		return NULL;
	}

	memset(in6, 0xFF, sizeof(struct in6_addr));

	for (i = 15; i > cidr/8-1; i--) {
		if (i == cidr/8) {
			in6->s6_addr[i] = in6->s6_addr[i] >> (8 - cidr % 8) << (8 - cidr %8);
		} else {
			in6->s6_addr[i] = 0;
		}
		// printf("ip6[%d]: %02x\n", i, in6->s6_addr[i]);
	}

	return in6;
}

struct sockaddr_storage* ss_apply_mask(struct sockaddr_storage* ip_ss, const struct sockaddr_storage* mask_ss) {
	union __sockaddr_union ip_u, mask_u;
	int i;

	if (ip_ss->ss_family != mask_ss->ss_family) {
		return NULL;
	}

	ip_u.ss = (struct sockaddr_storage*) ip_ss;
	mask_u.ss = (struct sockaddr_storage*) mask_ss;

	if (ip_ss->ss_family == AF_INET) {
		ip_u.sa_in->sin_addr.s_addr &= mask_u.sa_in->sin_addr.s_addr;
	} else if (ip_ss->ss_family == AF_INET6) {
		for (i = 15; i > 0 && mask_u.sa_in6->sin6_addr.s6_addr[i] != 0xFF; i--) {
			ip_u.sa_in6->sin6_addr.s6_addr[i] &= mask_u.sa_in6->sin6_addr.s6_addr[i];
		}
	}

	return ip_ss;
}

struct sockaddr_storage* ss_apply_cidr(struct sockaddr_storage* ip_ss, const unsigned int cidr) {
	union __sockaddr_union ip_u;
	int i;

	if ((ip_ss->ss_family == AF_INET && cidr > 32) || (ip_ss->ss_family == AF_INET6 && cidr > 128)) {
		return NULL;
	}

	ip_u.ss = (struct sockaddr_storage*) ip_ss;

	if (ip_ss->ss_family == AF_INET) {
		ip_u.sa_in->sin_addr.s_addr = ntohl(htonl(ip_u.sa_in->sin_addr.s_addr) >> (32 - cidr) << (32 - cidr));
	} else if (ip_ss->ss_family == AF_INET6) {
		for (i = 15; i > cidr / 8 - 1; i--) {
			if (i == cidr / 8) {
				ip_u.sa_in6->sin6_addr.s6_addr[i] = ip_u.sa_in6->sin6_addr.s6_addr[i] >> (8 - cidr % 8) << (8 - cidr % 8);
			} else {
				ip_u.sa_in6->sin6_addr.s6_addr[i] = 0;
			}
			// printf("ip6[%d]: %02x\n", i, in6->s6_addr[i]);
		}
	}

	return ip_ss;
}

struct in_addr* in_apply_cidr(struct in_addr* in, const unsigned int cidr) {
	if (cidr > 32) {
		return NULL;
	}

	in->s_addr = ntohl(htonl(in->s_addr) >> (32 - cidr) << (32 - cidr));

	return in;
}

struct in6_addr* in6_apply_cidr(struct in6_addr* in6, const unsigned int cidr) {
	int i;

	if (cidr > 128) {
		return NULL;
	}

	for (i = 15; i > cidr / 8 - 1; i--) {
		if (i == cidr / 8) {
			in6->s6_addr[i] = in6->s6_addr[i] >> (8 - cidr % 8) << (8 - cidr % 8);
		} else {
			in6->s6_addr[i] = 0;
		}
		// printf("ip6[%d]: %02x\n", i, in6->s6_addr[i]);
	}

	return in6;
}

#define HASH_ALGO_HYBRID /* Use the HYBRID hash algo. */
unsigned long hash_ip(struct sockaddr_storage *key) {
#if defined HASH_ALGO_HYBRID
	int c;
	char buf[46], *str;
	unsigned long k, hash = 5381;

	if (key->ss_family == AF_INET) { /* Very quick for IPv4 */
		k = ((struct sockaddr_in*) key)->sin_addr.s_addr;

		hash = (k<<16)^(k>>16)^k;
	} else {
		inet_ntop(key->ss_family, key, buf, sizeof(buf));
		str = buf;

		while ((c = *str++)) {
			hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
		}
	}

	return hash;
#elif defined HASH_ALGO_DJB2
	int c;
	char buf[46], *str;
	unsigned long hash = 5381;

	inet_ntop(key->ss_family, key, buf, sizeof(buf));
	str = buf;

	while (c = *str++) {
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
	}

	return hash;
#elif defined HASH_ALGO_SDBM
	/* This is a SDBM hash algo. */
	int c;
	char buf[46], *str;
	unsigned long hash = 0;

	inet_ntop(key->ss_family, key, buf, sizeof(buf));
	str = buf;

	while (c = *str++) {
		hash = c + (hash << 6) + (hash << 16) - hash;
	}

	return hash;
#elif defined HASH_ALGO_JENKINS
	/* This is a Bob Jenkins hash. */
	unsigned long hash, i;
	char buf[46];
	size_t len;

	inet_ntop(key->ss_family, key, buf, sizeof(buf));
	len = strlen(buf);

	for (hash = i = 0; i < len; ++i) {
		hash += buf[i];
		hash += (hash << 10);
		hash ^= (hash >> 6);
	}
	hash += (hash << 3);
	hash ^= (hash >> 11);
	hash += (hash << 15);

	return hash;
#else
	#error No hash algo selected.
#endif
}

#define HASH_ALGO_INT1
unsigned long hash(struct in_addr *key) {
#if defined HASH_ALGO_INT1
	unsigned long hash, k;

	k = key->s_addr;

	hash = ((k & 0xF0F0F0F0) >> 4) | ((k & 0x0F0F0F0F) << 4);

	return hash;
#elif defined HASH_ALGO_INT2
	unsigned long hash = 0, k;

	k = key->s_addr;

	hash = (k<<16)^(k>>16)^k;

	return hash;
#elif defined HASH_ALGO_INT3
	return key->s_addr;
#elif defined HASH_ALGO_DJB2
	int c;
	char buf[46], *str;
	unsigned long hash = 5381;

	inet_ntop(AF_INET, key, buf, sizeof(buf));
	str = buf;

	while (c = *str++) {
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
	}

	return hash;
#elif defined HASH_ALGO_SDBM
	/* This is a SDBM hash algo. */
	int c;
	char buf[46], *str;
	unsigned long hash = 0;

	inet_ntop(AF_INET, key, buf, sizeof(buf));
	str = buf;

	while (c = *str++) {
		hash = c + (hash << 6) + (hash << 16) - hash;
	}

	return hash;
#else
	#error No hash algo selected.
#endif
}

/*
** Local Variables:
** c-basic-offset: 3
** tab-width: 3
** End:
** vim: noet ts=3 sw=3
*/
